@@ -59,47 +59,55 @@ static void randomdelay(void)
 
 int sfgets(void)
 {
-    fd_set rs;
-    struct timeval tv;
-    ssize_t readen;
+    struct pollfd pfd;
+    int pollret;
+    ssize_t readnb;
     signed char seen_r = 0;
     static size_t scanned;
-    static size_t readend;
+    static size_t readnbd;
     
     if (scanned > (size_t) 0U) {       /* support pipelining */
-        readend -= scanned;        
-        memmove(cmd, cmd + scanned, readend);   /* safe */
+        readnbd -= scanned;        
+        memmove(cmd, cmd + scanned, readnbd);   /* safe */
         scanned = (size_t) 0U;
     }
-    tv.tv_sec = idletime;
-    tv.tv_usec = 0;        
-    FD_ZERO(&rs);
+    pfd.fd = clientfd;
+    pfd.events = POLLIN | POLLERR | POLLHUP;
     while (scanned < cmdsize) {
-        if (scanned >= readend) {      /* nothing left in the buffer */
-            FD_SET(0, &rs);
-            while (select(1, &rs, NULL, NULL, &tv) <= 0 && errno == EINTR);
-            if (FD_ISSET(0, &rs) == 0) {
+        if (scanned >= readnbd) {      /* nothing left in the buffer */
+            pfd.revents = 0;
+            while ((pollret = poll(&pfd, 1U, idletime * 1000UL)) < 0 &&
+                   errno == EINTR);
+            if (pollret == 0) {
                 return -1;
             }
-            if (readend >= cmdsize) {
+            if (pollret <= 0 ||
+                (pfd.revents & (POLLERR | POLLHUP | POLLNVAL)) != 0) {
+                return -2;
+            }
+            if ((pfd.revents & POLLIN) == 0) {
+                continue;
+            }
+            if (readnbd >= cmdsize) {
                 break;
             }
 #ifdef WITH_TLS
             if (tls_cnx != NULL) {
-                while ((readen = SSL_read
-                        (tls_cnx, cmd + readend, cmdsize - readend))
+                while ((readnb = SSL_read
+                        (tls_cnx, cmd + readnbd, cmdsize - readnbd))
                        < (ssize_t) 0 && errno == EINTR);
             } else
 #endif
             {
-                while ((readen = read(0, cmd + readend, cmdsize - readend))
-                       < (ssize_t) 0 && errno == EINTR);
+                while ((readnb = read(clientfd, cmd + readnbd,
+                                      cmdsize - readnbd)) < (ssize_t) 0 &&
+                       errno == EINTR);
             }
-            if (readen <= (ssize_t) 0) {
+            if (readnb <= (ssize_t) 0) {
                 return -2;
             }
-            readend += readen;
-            if (readend > cmdsize) {
+            readnbd += readnb;
+            if (readnbd > cmdsize) {
                 return -2;
             }
         }
@@ -116,8 +124,8 @@ int sfgets(void)
                     cmd[scanned] = 0;
                 }
 #endif
-                if (++scanned >= readend) {   /* non-pipelined command */
-                    scanned = readend = (size_t) 0U;
+                if (++scanned >= readnbd) {   /* non-pipelined command */
+                    scanned = readnbd = (size_t) 0U;
                 }
                 return 0;
             }
@@ -152,8 +160,8 @@ int sfgets(void)
 #else
 static char *revealextraspc(char * const s_)
 {
-    register unsigned char *s = (unsigned char *) s_;
-    register unsigned char *sn;
+    unsigned char *s = (unsigned char *) s_;
+    unsigned char *sn;
     
     if (s == NULL) {
         return s_;
@@ -181,19 +189,57 @@ static char *revealextraspc(char * const s_)
 }
 #endif
 
+#ifdef WITH_RFC2640
+char *charset_client2fs(const char * const string)
+{
+    char *output = NULL, *output_;
+    size_t inlen, outlen, outlen_;
+    
+    inlen = strlen(string);
+    outlen_ = outlen = inlen * (size_t) 4U + (size_t) 1U;
+    if (outlen <= inlen ||
+        (output_ = output = calloc(outlen, (size_t) 1U)) == NULL) {
+        die_mem();
+    }
+    if (utf8 > 0 && strcasecmp(charset_fs, "utf-8") != 0) {
+        if (iconv(iconv_fd_utf82fs, (char **) &string,
+                  &inlen, &output_, &outlen_) == (size_t) -1) {
+            strncpy(output, string, outlen);
+        }
+    } else if (utf8 <= 0 && strcasecmp(charset_fs, charset_client) != 0) {
+        if (iconv(iconv_fd_client2fs, (char **) &string,
+                  &inlen, &output_, &outlen_) == (size_t) -1) {
+            strncpy(output, string, outlen);
+        }
+    } else {
+        strncpy(output, string, outlen);
+    }
+    output[outlen - 1] = 0;    
+    
+    return output;
+}
+#endif
+
 void parser(void)
 {
     char *arg;
 #ifndef MINIMAL
     char *sitearg;
+#endif
+#ifdef WITH_RFC2640
+    char *narg = NULL;
 #endif
     size_t n;
 
+#ifdef IMPLICIT_TLS
+    (void) tls_init_new_session();
+    data_protection_level = CPL_PRIVATE;
+#endif
     for (;;) {
         xferfd = -1;
         if (state_needs_update != 0) {
             state_needs_update = 0;
-            setprogname("pure-ftpd (IDLE)");
+            setprocessname("pure-ftpd (IDLE)");
 #ifdef FTPWHO
             if (shm_data_cur != NULL) {
                 ftpwho_lock();
@@ -208,7 +254,7 @@ void parser(void)
         switch (sfgets()) {
         case -1:
 #ifdef BORING_MODE
-	    die(421, LOG_INFO, MSG_TIMEOUT);
+            die(421, LOG_INFO, MSG_TIMEOUT);
 #else
             die(421, LOG_INFO, MSG_TIMEOUT_PARSER);
 #endif
@@ -265,6 +311,10 @@ void parser(void)
                    cmd, strcmp(cmd, "pass") ? arg : "<*>");
 #endif
         }
+#ifdef WITH_RFC2640
+        narg = charset_client2fs(arg);
+	arg = narg;
+#endif
         /*
          * antiidle() is called with dummy commands, usually used by clients
          * who are wanting extra idle time. We give them some, but not too much.
@@ -273,7 +323,7 @@ void parser(void)
          */
         
 #ifndef MINIMAL
-        if (!strcmp(cmd, "noop") || !strcmp(cmd, "allo")) {
+        if (!strcmp(cmd, "noop")) {
             antiidle();
             donoop();
             goto wayout;
@@ -315,20 +365,36 @@ void parser(void)
             addreply_noformat(tls_cnx == NULL ? 503 : 200, "PBSZ=0");
         } else if (!strcmp(cmd, "prot")) {
             if (tls_cnx == NULL) {
-                addreply_noformat(503, "PBSZ?");
+                addreply_noformat(503, MSG_PROT_BEFORE_PBSZ);
                 goto wayout;
             }
             switch (*arg) {
             case 0:
                 addreply_noformat(503, MSG_MISSING_ARG);
+                data_protection_level = CPL_NONE;
                 break;
             case 'C':
                 if (arg[1] == 0) {
-                    addreply_noformat(200, "OK");
+                    addreply(200, MSG_PROT_OK, "clear");
+                    data_protection_level = CPL_CLEAR;
+                    break;
+                }
+            case 'S':
+            case 'E':
+                if (arg[1] == 0) {
+                    addreply(200, MSG_PROT_UNKNOWN_LEVEL, arg, "private");
+                    data_protection_level = CPL_PRIVATE;
+                    break;
+                }
+            case 'P':
+                if (arg[1] == 0) {
+                    addreply(200, MSG_PROT_OK, "private");
+                    data_protection_level = CPL_PRIVATE;
                     break;
                 }
             default:
                 addreply_noformat(534, "Fallback to [C]");
+                data_protection_level = CPL_CLEAR;
                 break;
             }
 #endif
@@ -346,6 +412,9 @@ void parser(void)
         } else if (!strcmp(cmd, "feat")) {
             dofeat();
             goto wayout;
+	} else if (!strcmp(cmd, "opts")) {
+	    doopts(arg);
+	    goto wayout;
 #endif
         } else if (!strcmp(cmd, "stru")) {
             dostru(arg);
@@ -374,12 +443,12 @@ void parser(void)
 #ifndef MINIMAL
             } else if (!strcmp(cmd, "eprt")) {
                 doeprt(arg);
-	    } else if (!strcmp(cmd, "esta") &&
-		       disallow_passive == 0 &&
-		       STORAGE_FAMILY(force_passive_ip) == 0) {
-		doesta();
-	    } else if (!strcmp(cmd, "estp")) {
-		doestp();
+            } else if (!strcmp(cmd, "esta") &&
+                       disallow_passive == 0 &&
+                       STORAGE_FAMILY(force_passive_ip) == 0) {
+                doesta();
+            } else if (!strcmp(cmd, "estp")) {
+                doestp();
 #endif
             } else if (disallow_passive == 0 && 
                        (!strcmp(cmd, "pasv") || !strcmp(cmd, "p@sw"))) {
@@ -399,16 +468,41 @@ void parser(void)
 #ifndef MINIMAL            
             } else if (disallow_passive == 0 && !strcmp(cmd, "spsv")) {
                 dopasv(2);
+            } else if (!strcmp(cmd, "allo")) {
+                if (*arg == 0) {
+                    addreply_noformat(501, MSG_STAT_FAILURE);
+                } else {
+                    const off_t size = (off_t) strtoull(arg, NULL, 10);
+                    doallo(size);
+                }
 #endif
             } else if (!strcmp(cmd, "pwd") || !strcmp(cmd, "xpwd")) {
+#ifdef WITH_RFC2640
+		char *nwd;
+#endif
                 antiidle();
+#ifdef WITH_RFC2640
+		nwd = charset_fs2client(wd);
+		addreply(257, "\"%s\" " MSG_IS_YOUR_CURRENT_LOCATION, nwd);
+		free(nwd);
+#else
                 addreply(257, "\"%s\" " MSG_IS_YOUR_CURRENT_LOCATION, wd);
+#endif
                 goto wayout;                
             } else if (!strcmp(cmd, "cdup") || !strcmp(cmd, "xcup")) {
                 docwd("..");
             } else if (!strcmp(cmd, "retr")) {
                 if (*arg != 0) {
-                    doretr(arg);
+#ifdef WITH_TLS
+                    if (enforce_tls_auth == 3 &&
+                        data_protection_level != CPL_PRIVATE) {
+                        addreply_noformat(521, MSG_PROT_PRIVATE_NEEDED);
+                    }
+                    else
+#endif
+                    {
+                        doretr(arg);
+                    }
                 } else {
                     addreply_noformat(501, MSG_NO_FILE_NAME);
                 }
@@ -430,20 +524,44 @@ void parser(void)
             } else if (!strcmp(cmd, "stor")) {
                 arg = revealextraspc(arg);
                 if (*arg != 0) {
-                    dostor(arg, 0, autorename);
+#ifdef WITH_TLS
+                    if (enforce_tls_auth == 3 &&
+                        data_protection_level != CPL_PRIVATE) {
+                        addreply_noformat(521, MSG_PROT_PRIVATE_NEEDED);
+                    } else 
+#endif
+                    {
+                        dostor(arg, 0, autorename);
+                    }
                 } else {
                     addreply_noformat(501, MSG_NO_FILE_NAME);
                 }
             } else if (!strcmp(cmd, "appe")) {
                 arg = revealextraspc(arg);
                 if (*arg != 0) {
-                    dostor(arg, 1, 0);
+#ifdef WITH_TLS
+                    if (enforce_tls_auth == 3 &&
+                        data_protection_level != CPL_PRIVATE) {
+                        addreply_noformat(521, MSG_PROT_PRIVATE_NEEDED);
+                    } else 
+#endif
+                    {
+                        dostor(arg, 1, 0);
+                    }
                 } else {
                     addreply_noformat(501, MSG_NO_FILE_NAME);
                 }
 #ifndef MINIMAL
             } else if (!strcmp(cmd, "stou")) {
-                dostou();
+#ifdef WITH_TLS
+                if (enforce_tls_auth == 3 &&
+                    data_protection_level != CPL_PRIVATE) {
+                    addreply_noformat(521, MSG_PROT_PRIVATE_NEEDED);
+                } else 
+#endif
+                {
+            	     dostou();
+                }
 #endif
 #ifndef DISABLE_MKD_RMD
             } else if (!strcmp(cmd, "mkd") || !strcmp(cmd, "xmkd")) {
@@ -463,15 +581,8 @@ void parser(void)
 #ifndef MINIMAL
             } else if (!strcmp(cmd, "stat")) {
                 if (*arg != 0) {
-# ifdef WITH_TLS
-                    if (tls_cnx != NULL) {
-                        addreply_noformat(500, MSG_UNKNOWN_COMMAND);
-                    } else
-# endif
-                    {
-                        modern_listings = 0;
-                        donlist(arg, 1, 1, 1);
-                    }
+                    modern_listings = 0;
+                    donlist(arg, 1, 1, 1, 1);
                 } else {
                     addreply_noformat(211, "http://www.pureftpd.org/");
                 }
@@ -480,18 +591,50 @@ void parser(void)
 #ifndef MINIMAL
                 modern_listings = 0;
 #endif
-                donlist(arg, 0, 1, 1);
+#ifdef WITH_TLS
+                if (enforce_tls_auth == 3 &&
+                    data_protection_level != CPL_PRIVATE) {
+                    addreply_noformat(521, MSG_PROT_PRIVATE_NEEDED);
+                } else
+#endif
+                {
+                    donlist(arg, 0, 1, 0, 1);
+                }
             } else if (!strcmp(cmd, "nlst")) {
 #ifndef MINIMAL                
                 modern_listings = 0;
 #endif
-                donlist(arg, 0, 0, broken_client_compat);
+#ifdef WITH_TLS
+                if (enforce_tls_auth == 3 &&
+                    data_protection_level != CPL_PRIVATE) {
+                    addreply_noformat(521, MSG_PROT_PRIVATE_NEEDED);
+                } else
+#endif
+                {
+                    donlist(arg, 0, 0, 0, broken_client_compat);
+                }
 #ifndef MINIMAL
             } else if (!strcmp(cmd, "mlst")) {
-                domlst(*arg != 0 ? arg : ".");
+#ifdef WITH_TLS
+                if (enforce_tls_auth == 3 &&
+                    data_protection_level != CPL_PRIVATE) {
+                    addreply_noformat(521, MSG_PROT_PRIVATE_NEEDED);
+                } else
+#endif
+                {
+                    domlst(*arg != 0 ? arg : ".");
+                }
             } else if (!strcmp(cmd, "mlsd")) {
                 modern_listings = 1;
-                donlist(arg, 0, 1, 0);
+#ifdef WITH_TLS
+                if (enforce_tls_auth == 3 &&
+                    data_protection_level != CPL_PRIVATE) {
+                    addreply_noformat(521, MSG_PROT_PRIVATE_NEEDED);
+                } else
+#endif
+                {
+                    donlist(arg, 0, 1, 1, 0);
+                }
 #endif
             } else if (!strcmp(cmd, "abor")) {
                 addreply_noformat(226, MSG_ABOR_SUCCESS);
@@ -531,7 +674,7 @@ void parser(void)
 # ifdef WITH_DIRALIASES
                                       " ALIAS" CRLF
 # endif
-                                      " CHMOD" CRLF " IDLE");
+                                      " CHMOD" CRLF " IDLE" CRLF " UTIME");
                     addreply_noformat(214, "Pure-FTPd - http://pureftpd.org/");
                 } else if (!strcasecmp(arg, "chmod")) {
                     char *sitearg2;
@@ -539,7 +682,7 @@ void parser(void)
                     
                     parsechmod:
                     if (sitearg == NULL || *sitearg == 0) {
-                        addreply_noformat(501, "SITE CHMOD: " MSG_MISSING_ARG);
+                        addreply_noformat(501, MSG_MISSING_ARG);
                         goto chmod_wayout;
                     }
                     sitearg2 = sitearg;
@@ -550,23 +693,63 @@ void parser(void)
                         sitearg2++;
                     }                    
                     if (*sitearg2 == 0) {
-                        addreply_noformat(550, "SITE CHMOD: " MSG_NO_FILE_NAME);
+                        addreply_noformat(550, MSG_NO_FILE_NAME);
                         goto chmod_wayout;
                     }
                     mode = (mode_t) strtoul(sitearg, NULL, 8);
                     if (mode > (mode_t) 07777) {
-                        addreply_noformat(501, "SITE CHMOD: " MSG_BAD_CHMOD);
+                        addreply_noformat(501, MSG_BAD_CHMOD);
                         goto chmod_wayout;
                     }
                     dochmod(sitearg2, mode);
-                  chmod_wayout:
+                    chmod_wayout:
                     (void) 0;
-# ifdef WITH_DIRALIASES
+                } else if (!strcasecmp(arg, "utime")) {
+                    char *sitearg2;
+                    
+                    if (sitearg == NULL || *sitearg == 0) {
+                        addreply_noformat(501, MSG_NO_FILE_NAME);
+                        goto utime_wayout;
+                    }		    
+                    if ((sitearg2 = strrchr(sitearg, ' ')) == NULL ||
+                        sitearg2 == sitearg) {
+                        addreply_noformat(501, MSG_MISSING_ARG);
+                        goto utime_wayout;
+                    }
+                    if (strcasecmp(sitearg2, " UTC") != 0) {
+                        addreply_noformat(500, "UTC Only");
+                        goto utime_wayout;			
+                    }
+                    *sitearg2-- = 0;
+                    if ((sitearg2 = strrchr(sitearg, ' ')) == NULL ||
+                        sitearg2 == sitearg) {
+                        utime_no_arg:
+                        addreply_noformat(501, MSG_MISSING_ARG);
+                        goto utime_wayout;
+                    }
+                    *sitearg2-- = 0;
+                    if ((sitearg2 = strrchr(sitearg, ' ')) == NULL ||
+                        sitearg2 == sitearg) {
+                        goto utime_no_arg;
+                    }
+                    *sitearg2-- = 0;
+                    if ((sitearg2 = strrchr(sitearg, ' ')) == NULL ||
+                        sitearg2 == sitearg) {
+                        goto utime_no_arg;
+                    }
+                    *sitearg2++ = 0;
+                    if (*sitearg2 == 0) {
+                        goto utime_no_arg;			
+                    }
+                    doutime(sitearg, sitearg2);
+                    utime_wayout:
+                    (void) 0;
+# ifdef WITH_DIRALIASES		    
                 } else if (!strcasecmp(arg, "alias")) {
                     if (sitearg == NULL || *sitearg == 0) {
                         print_aliases();
                     } else {
-                        register const char *alias;
+                        const char *alias;
                         
                         if ((alias = lookup_alias(sitearg)) != NULL) {
                             addreply(214, MSG_ALIASES_ALIAS, sitearg, alias);
@@ -608,7 +791,11 @@ void parser(void)
             }
         }
         noopidle = (time_t) -1;
-      wayout:
+        wayout:
+#ifdef WITH_RFC2640
+        free(narg);
+        narg = NULL;
+#endif
 #ifdef THROTTLING
         if (throttling_delay != 0UL) {
             usleep2(throttling_delay);
