@@ -8,12 +8,6 @@
 # define __attribute__(a)
 #endif
 
-#if defined(SYSTEM_QUOTAS) || defined(WITH_PRIVSEP)
-# ifdef HAVE_SYS_FSUID_H
-#  undef HAVE_SYS_FSUID_H
-# endif
-#endif
-
 #include <stdio.h>
 #ifdef STDC_HEADERS
 # include <stdlib.h>
@@ -67,6 +61,7 @@
 # include <sys/ioctl.h>
 #endif
 #include <sys/socket.h>
+#include <poll.h>
 #ifdef HAVE_NETINET_IN_SYSTM_H
 # include <netinet/in_systm.h>
 #endif
@@ -107,6 +102,9 @@
 #ifdef HAVE_UTIME_H
 # include <utime.h>
 #endif
+#ifdef WITH_RFC2640
+# include <iconv.h>
+#endif
 
 #include "mysnprintf.h"
 
@@ -114,7 +112,7 @@
 # define OLD_IP_STACK 1
 #endif
 
-#include "ipv4stack.h"
+#include "ipstack.h"
 
 /* We can't have more than one implementation, so if more than one were
  * found, the configure test failed - Use none */
@@ -123,22 +121,26 @@
 #  define DISABLE_SENDFILE 1
 # endif
 #endif
+/* Solaris now supports both sendfilev() and Linux-like sendfile().
+ * Prefer sendfile() if available. */
+#if defined(SENDFILEV_SOLARIS) && defined(SENDFILE_LINUX)
+# undef SENDFILEV_SOLARIS
+#endif
 #if defined(DISABLE_SENDFILE) || \
   (defined(SENDFILE_FREEBSD) && (defined(SENDFILE_LINUX) || defined(SENDFILEV_SOLARIS) || defined(SENDFILE_HPUX))) || \
   (defined(SENDFILE_LINUX) && (defined(SENDFILEV_SOLARIS) || defined(SENDFILE_HPUX))) || \
   (defined(SENDFILEV_SOLARIS) && defined(SENDFILE_HPUX))
 # undef SENDFILE_FREEBSD
 # undef SENDFILE_LINUX
+# undef SENDFILE64_LINUX
 # undef SENDFILE_HPUX
 # undef SENDFILEV_SOLARIS
 #endif
-#if defined(SENDFILE_FREEBSD) || defined(SENDFILEV_SOLARIS) || defined(SENDFILE_HPUX)
-/* FreeBSD, Solaris and HP-UX sendfile() handles large files */
-#elif defined(SENDFILE_LINUX)
-# ifdef WITH_LARGE_FILES
-#  undef SENDFILE_LINUX
-#  define SENDFILE_NONE
-# endif
+#if defined(SENDFILE_FREEBSD) || defined(SENDFILEV_SOLARIS) || \
+    defined(SENDFILE_HPUX) || \
+    (defined(SENDFILE_LINUX) && defined(SENDFILE64_LINUX))
+/* Old Linux kernels/glibcs that can't handle 64-bits sendfile() aren't
+ * supported any more. */
 #else
 # define SENDFILE_NONE
 #endif
@@ -157,9 +159,6 @@
 # include <sys/sendfile.h>
 #endif
 
-#ifdef HAVE_SYS_FSUID_H
-# include <sys/fsuid.h>
-#endif
 #ifdef HAVE_ALLOCA
 # ifdef HAVE_ALLOCA_H
 #  include <alloca.h>
@@ -265,6 +264,7 @@ typedef struct AuthResult_ {
     gid_t gid;
     const char *dir;
     int slow_tilde_expansion;
+    void *backend_data;
 #ifdef THROTTLING
     unsigned long throttling_bandwidth_ul;
     unsigned long throttling_bandwidth_dl;
@@ -288,7 +288,7 @@ typedef struct AuthResult_ {
 #endif
 } AuthResult;
 
-typedef struct FileInfo_ {
+typedef struct PureFileInfo_ {
     char **names_pnt;
     size_t name_offset;
     off_t size;
@@ -297,12 +297,12 @@ typedef struct FileInfo_ {
     nlink_t nlink;
     uid_t uid;
     gid_t gid;
-} FileInfo;
+} PureFileInfo;
 
 #define FI_NAME(X) (*((X)->names_pnt) + (X)->name_offset)
 
 typedef enum {
-    ALTLOG_NONE, ALTLOG_CLF, ALTLOG_STATS, ALTLOG_W3C
+    ALTLOG_NONE, ALTLOG_CLF, ALTLOG_STATS, ALTLOG_W3C, ALTLOG_XFERLOG
 } AltLogFormat;
 
 typedef struct AltLogPrefixes_ {
@@ -310,8 +310,37 @@ typedef struct AltLogPrefixes_ {
     const AltLogFormat format;
 } AltLogPrefixes;
 
+typedef enum {
+    CPL_NONE, CPL_CLEAR, CPL_SAFE, CPL_CONFIDENTIAL, CPL_PRIVATE
+} ChannelProtectionLevel;
+
+int pureftpd_start(int argc, char *argv[], const char *home_directory);
+
+#ifdef __IPHONE__
+void pureftpd_register_login_callback(void (*callback)(void *user_data),
+                                      void *user_data);
+
+void pureftpd_register_logout_callback(void (*callback)(void *user_data),
+                                       void *user_data);
+
+void pureftpd_register_log_callback(void (*callback)(int crit,
+                                                     const char *message,
+                                                     void *user_data),
+                                    void *user_data);
+
+void pureftpd_register_simple_auth_callback(int (*callback)(const char *account,
+                                                            const char *password,
+                                                            void *user_data),
+                                            void *user_data);
+
+int pureftpd_shutdown(void);
+int pureftpd_enable(void);
+int pureftpd_disable(void);
+#endif
 int safe_write(const int fd, const void *buf_, size_t count);
-void *aborttransfer(int);
+#ifdef WITH_TLS
+int secure_safe_write(void * const tls_fd, const void *buf_, size_t count);
+#endif
 void parser(void);
 void stripctrl(char * const buf, size_t len);
 void dobanner(const int type);
@@ -321,6 +350,7 @@ void docwd(const char *dir);
 void doretr(char *name);
 void dorest(const char *name);
 void dodele(char *name);
+void doallo(off_t size);
 void dostor(char *name, const int append, const int autorename);
 void domkd(char *name);
 void dormd(char *name);
@@ -332,9 +362,12 @@ void doport2(struct sockaddr_storage a, unsigned int p);
 #ifndef MINIMAL
 void doesta(void);
 void doestp(void);
+void doallo(const off_t size);
 #endif
 void dopasv(int);
+void doopts(char *args);
 void dochmod(char *name, mode_t mode);
+void doutime(char *name, const char * const wanted_time);
 void error(int n, const char *msg);
 void domode(const char *arg);
 void dostru(const char *arg);
@@ -346,30 +379,34 @@ void dostou(void);
 void dofeat(void);
 void domlst(const char * const file);
 void dositetime(void);
+int ul_check_free_space(const char *name);
 void mappedtov4(struct sockaddr_storage *ss);
-#ifndef HAVE_SYS_FSUID_H
 void disablesignals(void);
-#endif
 void getnames(void);
 void donlist(char *arg, const int on_ctrlconn, const int opt_l_,
-             const int split_args);
+             const int opt_a_, const int split_args);
 void opendata(void);
 void closedata(void);
+void client_fflush(void);
+void client_printf(const char * const format, ...)
+    __attribute__ ((format(printf, 1, 2)));    
 void addreply(const int code, const char * const line, ...)
-	__attribute__ ((format(printf, 2, 3)));
+    __attribute__ ((format(printf, 2, 3)));
 void addreply_noformat(const int code, const char * const line);
 void doreply(void);
 void sighandler(int sig);
 void prevent(char *arg);
-unsigned int daemons(unsigned short server_port);
+unsigned int daemons(in_port_t server_port);
 void logfile(const int facility, const char *format, ...) 
-	__attribute__ ((format(printf, 2, 3)));
+    __attribute__ ((format(printf, 2, 3)));
+char *skip_telnet_controls(const char *str);
 void die(const int err, const int priority, const char * const format, ...)
-	__attribute__ ((format(printf, 3, 4)));
-void die_mem(void);
-void setprogname(const char * const title);
-int modernformat(const char *file,
-                 char *target, size_t target_size);
+    __attribute__ ((format(printf, 3, 4))) __attribute__ ((noreturn));
+void die_mem(void) __attribute__ ((noreturn));
+void _EXIT(const int status) __attribute__ ((noreturn));
+void setprocessname(const char * const title);
+int modernformat(const char *file, char *target, size_t target_size,
+                 const char * const prefix);
 int sfgets(void);
 const char *getgroup(const gid_t gid);
 const char *getname(const uid_t uid);
@@ -377,13 +414,10 @@ unsigned int zrand(void);
 void simplify(char *subdir);
 int checkprintable(register const char *s);
 void delete_atomic_file(void);
-
-#ifdef HAVE_SYS_FSUID_H
-# define usleep2 usleep
-#else
-void usleep2(const unsigned long microsec);
+#ifdef WITH_RFC2640
+char *charset_fs2client(const char * const string);
 #endif
-
+void usleep2(const unsigned long microsec);
 int mysnprintf(char *str, size_t size, const char *format, ...);
 
 extern int opt_a, opt_C, opt_d, opt_F, opt_l, opt_R;
@@ -408,7 +442,7 @@ extern int opt_a, opt_C, opt_d, opt_F, opt_l, opt_R;
 # ifdef NON_ROOT_FTP
 #  define PID_FILE CONFDIR "/pure-ftpd.pid"
 # else
-#  define PID_FILE "/var/run/pure-ftpd.pid"
+#  define PID_FILE STATEDIR "/run/pure-ftpd.pid"
 # endif
 #endif
 
@@ -416,7 +450,7 @@ extern int opt_a, opt_C, opt_d, opt_F, opt_l, opt_R;
 # ifdef NON_ROOT_FTP
 #  define UPLOADSCRIPT_PID_FILE CONFDIR "/pure-uploadscript.pid"
 # else
-#  define UPLOADSCRIPT_PID_FILE "/var/run/pure-uploadscript.pid"
+#  define UPLOADSCRIPT_PID_FILE STATEDIR "/run/pure-uploadscript.pid"
 # endif
 #endif
 
@@ -424,16 +458,28 @@ extern int opt_a, opt_C, opt_d, opt_F, opt_l, opt_R;
 # ifdef NON_ROOT_FTP
 #  define AUTHD_PID_FILE CONFDIR "/pure-authd.pid"
 # else
-#  define AUTHD_PID_FILE "/var/run/pure-authd.pid"
+#  define AUTHD_PID_FILE STATEDIR "/run/pure-authd.pid"
 # endif
 #endif
 
 #ifndef NON_ROOT_FTP
-# define DEFAULT_FTP_PORT_S "21"
+# ifdef IMPLICIT_TLS
+#  define DEFAULT_FTP_PORT_S "990"
+# else
+#  define DEFAULT_FTP_PORT_S "21"
+# endif
 #elif defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
-# define DEFAULT_FTP_PORT_S "21"
+# ifdef IMPLICIT_TLS
+#  define DEFAULT_FTP_PORT_S "990"
+# else
+#  define DEFAULT_FTP_PORT_S "21"
+# endif
 #else
-# define DEFAULT_FTP_PORT_S "2121"
+# ifdef IMPLICIT_TLS
+#  define DEFAULT_FTP_PORT_S "9990"
+# else
+#  define DEFAULT_FTP_PORT_S "2121"
+# endif
 #endif
 
 /*
@@ -463,17 +509,46 @@ the server may be insecure if a wrong value is set here.
 Your platform has a very large MAXPATHLEN, we should not trust it.
 #endif
 
-#define DEFAULT_MAX_USERS 50
-#define DEFAULT_FTP_DATA_PORT 20 
-#define MAX_SYSLOG_LINE (MAXPATHLEN + 512U)
-#define DEFAULT_IDLE (15UL * 60UL)
-#define MAX_SITE_IDLE (42UL * 60UL)
-#define DEFAULT_MAX_LS_FILES 2000U
-#define DEFAULT_MAX_LS_DEPTH 5U
-#define GLOB_TIMEOUT 17                   /* Max user time for a 'ls' to complete */
-#define MAX_CPU_TIME (17 * 60)           /* Max allowed CPU time per session */
-#define MAX_SESSION_XFER_IDLE (24 * 60 * 60)   /* Max duration of a transfer */
-#define MAX_USER_LENGTH 32U
+#ifdef __IPHONE__
+# define DEFAULT_MAX_USERS 1
+#endif
+#ifndef DEFAULT_MAX_USERS    
+# define DEFAULT_MAX_USERS 50
+#endif
+#ifndef DEFAULT_FTP_DATA_PORT
+# ifdef IMPLICIT_TLS
+#  define DEFAULT_FTP_DATA_PORT 989
+# else
+#  define DEFAULT_FTP_DATA_PORT 20
+# endif
+#endif
+#ifndef MAX_SYSLOG_LINE
+# define MAX_SYSLOG_LINE (MAXPATHLEN + 512U)
+#endif
+#ifndef DEFAULT_IDLE
+# define DEFAULT_IDLE (15UL * 60UL)
+#endif
+#ifndef MAX_SITE_IDLE
+# define MAX_SITE_IDLE (42UL * 60UL)
+#endif
+#ifndef DEFAULT_MAX_LS_FILES    
+# define DEFAULT_MAX_LS_FILES 10000U
+#endif
+#ifndef DEFAULT_MAX_LS_DEPTH
+# define DEFAULT_MAX_LS_DEPTH 5U
+#endif
+#ifndef GLOB_TIMEOUT
+# define GLOB_TIMEOUT 17                   /* Max user time for a 'ls' to complete */
+#endif
+#ifndef MAX_CPU_TIME
+# define MAX_CPU_TIME (4 * 60 * 60)           /* Max allowed CPU time per session */
+#endif
+#ifndef MAX_SESSION_XFER_IDLE    
+# define MAX_SESSION_XFER_IDLE (24 * 60 * 60)   /* Max duration of a transfer */
+#endif
+#ifndef MAX_USER_LENGTH
+# define MAX_USER_LENGTH 127U
+#endif
 
 #ifdef LOG_FTP
 # define DEFAULT_FACILITY LOG_FTP
@@ -496,38 +571,55 @@ Your platform has a very large MAXPATHLEN, we should not trust it.
 # define CONF_TCP_SO_SNDBUF 65536
 #endif
     
-#define DEFAULT_DL_CHUNK_SIZE (1 << 18)
-#define DEFAULT_UL_CHUNK_SIZE (CONF_TCP_SO_RCVBUF)
-#define MAX_UL_CHUNK_SIZE (CONF_TCP_SO_RCVBUF)
-#define MAX_THROTTLING_DELAY 42           /* Maximum throttling compensation */
+#ifndef DL_MIN_CHUNK_SIZE
+# define DL_MIN_CHUNK_SIZE (8 * 1024)
+#endif
+#ifndef DL_DEFAULT_CHUNK_SIZE
+# define DL_DEFAULT_CHUNK_SIZE 49152UL
+#endif
+#ifndef DL_DEFAULT_CHUNK_SIZE_ASCII
+# define DL_DEFAULT_CHUNK_SIZE_ASCII 32768UL
+#endif
+#ifndef DL_MAX_CHUNK_SIZE
+# define DL_MAX_CHUNK_SIZE (128 * 1024UL)
+#endif
+#ifndef DL_DLMAP_SIZE
+# define DL_DLMAP_SIZE (128 * 1024UL)
+#endif
+#if DL_DEFAULT_CHUNK_SIZE > DL_MAX_CHUNK_SIZE || DL_MIN_CHUNK_SIZE > DL_MAX_CHUNK_SIZE
+# error DL_MAX_CHUNK_SIZE shouldnt be <= DL_MIN_CHUNK_SIZE or <= DL_DEFAULT_CHUNK_SIZE
+#endif
+#if DL_DLMAP_SIZE < DL_MAX_CHUNK_SIZE
+# error DL_DLMAP_SIZE should be >= DL_MAX_CHUNK_SIZE
+#endif
+
+#ifndef UL_MIN_CHUNK_SIZE
+# define UL_MIN_CHUNK_SIZE CONF_TCP_SO_RCVBUF
+#endif
+#ifndef UL_DEFAULT_CHUNK_SIZE
+# define UL_DEFAULT_CHUNK_SIZE (CONF_TCP_SO_RCVBUF * 2)
+#endif
+#ifndef UL_DEFAULT_CHUNK_SIZE_ASCII
+# if CONF_TCP_SO_RCVBUF < 65536
+#  define UL_DEFAULT_CHUNK_SIZE_ASCII CONF_TCP_SO_RCVBUF
+# else
+#  define UL_DEFAULT_CHUNK_SIZE_ASCII 65536UL
+# endif
+#endif
+#ifndef UL_MAX_CHUNK_SIZE
+# define UL_MAX_CHUNK_SIZE (512 * 1024UL)
+#endif    
 
 #define VHOST_PREFIX_MAX_LEN 64    
 
 #define PUREFTPD_TMPFILE_PREFIX ".pureftpd-"    
 #define ATOMIC_PREFIX_PREFIX PUREFTPD_TMPFILE_PREFIX "upload."
-    
-/*
- * Some users reported that something was wrong with TCP_CORK.
- * Better disable it for now, but we have to investigate.
- */
-#ifdef TCP_CORK
-# undef TCP_CORK
-#endif
-#ifdef TCP_CORK
-# define CORK_ON(SK) do { int optval = 1; setsockopt(SK, SOL_TCP, TCP_CORK, \
-  &optval, sizeof optval); } while(0)
-# define CORK_OFF(SK) do { int optval = 0; setsockopt(SK, SOL_TCP, TCP_CORK, \
-  &optval, sizeof optval); } while(0)
-#else
-# define CORK_ON(SK) do { } while(0)
-# define CORK_OFF(SK) do { } while(0)
-#endif
 
-#define STORAGE_PORT(X) (((struct sockaddr_in *) &(X))->sin_port)
-#define STORAGE_PORT6(X) (((struct sockaddr_in6 *) &(X))->sin6_port)
-#define STORAGE_SIN_ADDR(X) ((((struct sockaddr_in *) &(X))->sin_addr).s_addr)
-#define STORAGE_SIN_ADDR6(X) ((((struct sockaddr_in6 *) &(X))->sin6_addr).s6_addr)
-#define STORAGE_SIN_ADDR6_NF(X) (((struct sockaddr_in6 *) &(X))->sin6_addr)
+#define STORAGE_PORT(X)  (*storage_port(&(X)))
+#define STORAGE_PORT6(X) (*storage_port6(&(X)))
+#define STORAGE_SIN_ADDR(X) (storage_sin_addr(&(X))->s_addr)
+#define STORAGE_SIN_ADDR6(X) (storage_sin_addr6(&(X))->s6_addr)
+#define STORAGE_SIN_ADDR6_NF(X) (*(storage_sin_addr6(&(X))))
 
 #ifdef HAVE_SS_LEN
 # define STORAGE_LEN(X) ((X).ss_len)
@@ -653,10 +745,10 @@ Your platform has a very large MAXPATHLEN, we should not trust it.
 # define S_IXUGO (S_IXUSR | S_IXGRP | S_IXOTH)
 #endif
 
-#ifdef ACCEPT_UNICODE_CONTROL_CHARS
-# define ISCTRLCODE(X) ((X) == 0x7f || ((unsigned char) (X)) < 32U)
-#else
+#ifdef DISABLE_UNICODE_CONTROL_CHARS
 # define ISCTRLCODE(X) ((X) == 0x7f || !(((unsigned char) (X)) & 0x60))
+#else
+# define ISCTRLCODE(X) ((X) == 0x7f || ((unsigned char) (X)) < 32U)
 #endif
     
 #ifndef HAVE_MUNMAP
@@ -692,12 +784,10 @@ Your platform has a very large MAXPATHLEN, we should not trust it.
 # define _exit(X) exit(X)
 #endif
 
-#ifdef FTPWHO
-# define _EXIT(X) do { delete_atomic_file(); ftpwho_exit(X); } while(0)
-#else
-# define _EXIT(X) do { delete_atomic_file(); _exit(X); } while(0)
-#endif
-
+#ifndef offsetof
+# define offsetof(type, member) ((size_t) &((type*) NULL)->member)
+#endif    
+    
 #include "bsd-realpath.h"    
 #include "fakechroot.h"
         
