@@ -7,12 +7,15 @@
 #include "dynamic.h"
 #include "ftpwho-update.h"
 #include "globals.h"
+#ifdef WITH_TLS
+# include "tls.h"
+#endif
 
 #ifdef WITH_DMALLOC
 # include <dmalloc.h>
 #endif
 
-static void wrstr(const int f, const char *s)
+static void wrstr(const int f, void * const tls_fd, const char *s)
 {
     static char outbuf[CONF_TCP_SO_SNDBUF];
     static size_t outcnt;
@@ -20,8 +23,18 @@ static void wrstr(const int f, const char *s)
     
     if (s == NULL) {
         if (outcnt > (size_t) 0U) {
-            if (safe_write(f, outbuf, outcnt) != 0) {
-                return;
+#ifdef WITH_TLS
+            if (tls_fd != NULL) {
+                if (secure_safe_write(tls_fd, outbuf, outcnt) < 0) {
+                    return;
+                } 
+            } else
+#endif      
+            {
+                (void) tls_fd;
+                if (safe_write(f, outbuf, outcnt) != 0) {
+                    return;
+                }
             }
         }
         outcnt = (size_t) 0U;
@@ -42,15 +55,37 @@ static void wrstr(const int f, const char *s)
         s += rest;
         l -= rest;
     }
-    if (safe_write(f, outbuf, sizeof outbuf) != 0) {
-        return;
-    }
-    while (l > sizeof outbuf) {
-        if (safe_write(f, s, sizeof outbuf) != 0) {
+#ifdef WITH_TLS
+    if (data_protection_level == CPL_PRIVATE) {
+        if (secure_safe_write(tls_fd, outbuf, sizeof outbuf) < 0) {
+            return;
+        } 
+    } else
+#endif
+    {       
+        if (safe_write(f, outbuf, sizeof outbuf) != 0) {
             return;
         }
-        s += sizeof outbuf;
-        l -= sizeof outbuf;
+    }
+#ifdef WITH_TLS
+    if (data_protection_level == CPL_PRIVATE) {    
+        while (l > sizeof outbuf) {
+            if (secure_safe_write(tls_fd, s, sizeof outbuf) < 0) {
+                return;
+            }
+            s += sizeof outbuf;
+            l -= sizeof outbuf;
+        } 
+    } else
+#endif
+    {
+        while (l > sizeof outbuf) {
+            if (safe_write(f, s, sizeof outbuf) != 0) {
+                return;
+            }
+            s += sizeof outbuf;
+            l -= sizeof outbuf;
+        }
     }
     if (l > (size_t) 0U) {
         memcpy(outbuf, s, l);          /* safe, l <= sizeof outbuf */
@@ -61,17 +96,17 @@ static void wrstr(const int f, const char *s)
 #ifdef NO_FTP_USERS
 const char *getname(const uid_t uid)
 {
-    static char number[9];
+    static char number[11];
 
-    snprintf(number, sizeof number, "%-8d", uid);
+    snprintf(number, sizeof number, "%-10d", uid);
     return number;
 }
 
 const char *getgroup(const gid_t gid)
 {
-    static char number[9];
+    static char number[11];
 
-    snprintf(number, sizeof number, "%-8d", gid);
+    snprintf(number, sizeof number, "%-10d", gid);
     return number;
 }
 #else
@@ -97,17 +132,17 @@ const char *getname(const uid_t uid)
         die_mem();
     }
     p->uid = uid;
-    if ((p->name = malloc((size_t) 9U)) == NULL) {
+    if ((p->name = malloc((size_t) 11U)) == NULL) {
         die_mem();
     }
     if (pwd != NULL) {
-        if (SNCHECK(snprintf(p->name, (size_t) 9U, 
-                             "%-8.8s", pwd->pw_name), (size_t) 9U)) {
+        if (SNCHECK(snprintf(p->name, (size_t) 11U, 
+                             "%-10.10s", pwd->pw_name), (size_t) 11U)) {
             _EXIT(EXIT_FAILURE);
         }
     } else {
-        if (SNCHECK(snprintf(p->name, (size_t) 9U, "%-8d", uid), 
-                    (size_t) 9U)) {
+        if (SNCHECK(snprintf(p->name, (size_t) 11U, "%-10d", uid), 
+                    (size_t) 11U)) {
             _EXIT(EXIT_FAILURE);
         }
     }
@@ -138,17 +173,17 @@ const char *getgroup(const gid_t gid)
         die_mem();
     }
     p->gid = gid;
-    if ((p->name = malloc((size_t) 9U)) == NULL) {
+    if ((p->name = malloc((size_t) 11U)) == NULL) {
         die_mem();
     }
     if (pwd != NULL) {
-        if (SNCHECK(snprintf(p->name, (size_t) 9U, "%-8.8s",
-                             pwd->gr_name), (size_t) 9U)) {
+        if (SNCHECK(snprintf(p->name, (size_t) 11U, "%-10.10s",
+                             pwd->gr_name), (size_t) 11U)) {
             _EXIT(EXIT_FAILURE);
         }
     } else {
-        if (SNCHECK(snprintf(p->name, (size_t) 9U, "%-8d", gid), 
-                    (size_t) 9U)) {
+        if (SNCHECK(snprintf(p->name, (size_t) 11U, "%-10d", gid), 
+                    (size_t) 11U)) {
             _EXIT(EXIT_FAILURE);
         }
     }
@@ -213,7 +248,8 @@ static int listfile(const PureFileInfo * const fi,  const char *name)
         if ((alloca_nameline = ALLOCA(sizeof_nameline)) == NULL) {
             return 0;
         }
-        if ((rval = modernformat(n, alloca_nameline, sizeof_nameline)) < 0) {
+        if ((rval = modernformat(n, alloca_nameline,
+                                 sizeof_nameline, "")) < 0) {
             ALLOCA_FREE(alloca_nameline);            
             return 0;
         }
@@ -237,10 +273,24 @@ static int listfile(const PureFileInfo * const fi,  const char *name)
         name = FI_NAME(fi);
     }
 #if defined(WITH_VIRTUAL_CHROOT) && defined(S_IFLNK) && defined(S_IFDIR)
-    if (S_ISLNK(st.st_mode) && name[0] == '.' && name[1] == '.' && name[2] == 0) {
+    if (S_ISLNK(st.st_mode) && name[0] == '.' &&
+        name[1] == '.' && name[2] == 0) {
         st.st_mode &= ~S_IFLNK;
         st.st_mode |= S_IFDIR;
-    }  /* Hack to please some Windows client that dislike ../ -> ../ */
+    }  /* Hack to please some Windows clients that dislike ../ -> ../ */
+#endif
+#if !defined(MINIMAL) && !defined(ALWAYS_SHOW_SYMLINKS_AS_SYMLINKS)
+    if (
+# ifndef ALWAYS_SHOW_RESOLVED_SYMLINKS
+        broken_client_compat != 0 &&
+# endif
+        S_ISLNK(st.st_mode)) {
+        struct stat sts;
+        
+        if (stat(name, &sts) == 0 && !S_ISLNK(sts.st_mode)) {
+            st = sts;
+        }
+    } /* Show non-dangling symlinks as files/directories */
 #endif
 #ifdef DISPLAY_FILES_IN_UTC_TIME
     t = gmtime((time_t *) &st.st_mtime);
@@ -328,36 +378,22 @@ static int listfile(const PureFileInfo * const fi,  const char *name)
             if ((alloca_nameline = ALLOCA(sizeof_nameline)) == NULL) {
                 return 0;
             }
-#ifdef WITH_LARGE_FILES
             if (SNCHECK(snprintf(alloca_nameline, sizeof_nameline,
-                                 "%s %4u %s %s %8llu %s %2d %s %s", 
+                                 "%s %4u %s %s %10llu %s %2d %s %s", 
                                  m, (unsigned int) st.st_nlink,
-				 getname(st.st_uid),
+                                 getname(st.st_uid),
                                  getgroup(st.st_gid), 
                                  (unsigned long long) st.st_size,
                                  months[t->tm_mon],
                                  t->tm_mday, timeline, name),
-                        sizeof_nameline))
-#else
-            if (SNCHECK(snprintf(alloca_nameline, sizeof_nameline,
-                                 "%s %4u %s %s %8lu %s %2d %s %s", 
-                                 m, (unsigned int) st.st_nlink,
-				 getname(st.st_uid),
-                                 getgroup(st.st_gid), 
-                                 (unsigned long) st.st_size,
-                                 months[t->tm_mon],
-                                 t->tm_mday, timeline, name),
-                        sizeof_nameline))
-#endif
-            {
+                        sizeof_nameline)) {
                 ALLOCA_FREE(alloca_nameline);
                 _EXIT(EXIT_FAILURE);
             }
             if (S_ISLNK(st.st_mode)) {
-                char *p = alloca_nameline + strlen(alloca_nameline);
-                
+                char *p = alloca_nameline + strlen(alloca_nameline);                
                 {
-                    int sx;
+                    ssize_t sx;
                     
                     if ((sx = readlink(name, m, sizeof m - 1U)) > 0) {
                         m[sx] = 0;
@@ -397,15 +433,14 @@ static int listfile(const PureFileInfo * const fi,  const char *name)
     return rval;
 }
 
-static void outputfiles(int f)
+static void outputfiles(int f, void * const tls_fd)
 {
     unsigned int n;
     struct filename *p;
     struct filename *q;
-#ifdef WITH_RFC2640
     char *c_buf; /* buffer with charset of client */
-#endif
 
+    (void) c_buf;
     if (!head) {
         return;
     }
@@ -474,13 +509,13 @@ static void outputfiles(int f)
                 pad[2] = 0;
             }
 #ifdef WITH_RFC2640
-	    c_buf = charset_fs2client(q->line);
-            wrstr(f, c_buf);
-	    free(c_buf);
+            c_buf = charset_fs2client(q->line);
+            wrstr(f, tls_fd, c_buf);
+            free(c_buf);
 #else
-            wrstr(f, q->line);
+            wrstr(f, tls_fd, q->line);
 #endif
-            wrstr(f, pad);
+            wrstr(f, tls_fd, pad);
             q = q->right;
             free(tmp);
         }
@@ -639,12 +674,14 @@ static PureFileInfo *sreaddir(char **names_pnt)
 }
 
 /* have to change to the directory first (speed hack for -R) */
-static void listdir(unsigned int depth, int f, const char *name)
+static void listdir(unsigned int depth, int f, void * const tls_fd,
+                    const char *name)
 {
     PureFileInfo *dir;
     char *names;
     PureFileInfo *s;
     PureFileInfo *r;
+    char *c_buf;
     int d;
     
     if (depth >= max_ls_depth || matches >= max_ls_files) {
@@ -653,7 +690,7 @@ static void listdir(unsigned int depth, int f, const char *name)
     if ((dir = sreaddir(&names)) == NULL) {
         addreply(226, MSG_CANT_READ_FILE, name);
         return;
-    }        
+    }
     s = dir;
     while (s->name_offset != (size_t) -1) {
         if (FI_NAME(s)[0] != '.') {
@@ -672,7 +709,7 @@ static void listdir(unsigned int depth, int f, const char *name)
         }
         s++;
     }
-    outputfiles(f);
+    outputfiles(f, tls_fd);
     r = dir;
     while (opt_R && r != s) {
         if (r->name_offset != (size_t) -1 && !chdir(FI_NAME(r))) {
@@ -686,22 +723,30 @@ static void listdir(unsigned int depth, int f, const char *name)
                                  name, FI_NAME(r)), sizeof_subdir)) {
                 goto nolist;
             }
+#ifdef WITH_RFC2640
+            c_buf = charset_fs2client(alloca_subdir);
+#else
+            c_buf = alloca_subdir;
+#endif
 #ifndef MINIMAL
             if (modern_listings == 0) {
-#endif
+#endif                
 #ifdef FANCY_LS_DIRECTORY_HEADERS
-                wrstr(f, "\r\n>----------------[");
-                wrstr(f, alloca_subdir);
-                wrstr(f, "]----------------<\r\n\r\n");
+                wrstr(f, tls_fd, "\r\n>----------------[");
+                wrstr(f, tls_fd, c_buf);
+                wrstr(f, tls_fd, "]----------------<\r\n\r\n");
 #else
-                wrstr(f, "\r\n\r\n");
-                wrstr(f, alloca_subdir);
-                wrstr(f, ":\r\n\r\n");                    
+                wrstr(f, tls_fd, "\r\n\r\n");
+                wrstr(f, tls_fd, c_buf);
+                wrstr(f, tls_fd, ":\r\n\r\n");                    
 #endif
 #ifndef MINIMAL
             }
 #endif
-            listdir(depth + 1U, f, alloca_subdir);
+#ifdef WITH_RFC2640            
+            free(c_buf);
+#endif            
+            listdir(depth + 1U, f, tls_fd, alloca_subdir);
             nolist:
             ALLOCA_FREE(alloca_subdir);
             if (matches >= max_ls_files) {
@@ -724,10 +769,12 @@ static void listdir(unsigned int depth, int f, const char *name)
 }
 
 void donlist(char *arg, const int on_ctrl_conn, const int opt_l_,
-             const int split_args)
+             const int opt_a_, const int split_args)
 {
     int c;
-
+    void *tls_fd = NULL;
+    char *c_buf;
+    
     matches = 0U;
 
     opt_C = opt_d = opt_F = opt_R = opt_r = opt_t = opt_S = 0;
@@ -735,7 +782,7 @@ void donlist(char *arg, const int on_ctrl_conn, const int opt_l_,
     if (force_ls_a != 0) {
         opt_a = 1;
     } else {
-        opt_a = 0;
+        opt_a = opt_a_;
     }
     if (split_args != 0) {
         while (isspace((unsigned char) *arg)) {
@@ -786,7 +833,7 @@ void donlist(char *arg, const int on_ctrl_conn, const int opt_l_,
         }
     }
     if (on_ctrl_conn == 0) {
-    opendata();
+        opendata();
         if ((c = xferfd) == -1) {
             return;
         }
@@ -794,10 +841,25 @@ void donlist(char *arg, const int on_ctrl_conn, const int opt_l_,
         if (type == 2) {
             addreply_noformat(0, "ASCII");
         }
+#ifdef WITH_TLS
+        if (data_protection_level == CPL_PRIVATE) {
+            tls_fd = tls_data_cnx;
+        }
+#endif
     } else {                           /* STAT command */
-        c = 1;
-        safe_write(c, "213-STAT" CRLF, sizeof "213-STAT" CRLF - 1U);
-    }        
+        c = clientfd;
+#ifdef WITH_TLS
+        if (data_protection_level == CPL_PRIVATE) {
+            secure_safe_write(tls_cnx, "213-STAT" CRLF,
+                              sizeof "213-STAT" CRLF - 1U);
+            tls_fd = tls_cnx;
+        }
+        else
+#endif
+        {
+            safe_write(c, "213-STAT" CRLF, sizeof "213-STAT" CRLF - 1U);
+        }
+    }
     if (arg != NULL && *arg != 0) {
         int justone;
 
@@ -847,7 +909,7 @@ void donlist(char *arg, const int on_ctrl_conn, const int opt_l_,
                     }
                     path++;
                 }
-                outputfiles(c);    /* in case of opt_C */
+                outputfiles(c, tls_fd);    /* in case of opt_C */
                 path = g.gl_pathv;
                 while (path && *path) {
                     if (matches >= max_ls_files) {
@@ -855,18 +917,26 @@ void donlist(char *arg, const int on_ctrl_conn, const int opt_l_,
                     }
                     if (**path) {
                         if (!justone) {
-#ifdef FANCY_LS_DIRECTORY_HEADERS                            
-                            wrstr(c, "\r\n>-----------------[");
-                            wrstr(c, *path);
-                            wrstr(c, "]-----------------<\r\n\r\n");
+#ifdef WITH_RFC2640                            
+                            c_buf = charset_fs2client(*path);
 #else
-                            wrstr(c, "\r\n\r\n");
-                            wrstr(c, *path);
-                            wrstr(c, ":\r\n\r\n");
+                            c_buf = *path;
+#endif
+#ifdef FANCY_LS_DIRECTORY_HEADERS                            
+                            wrstr(c, tls_fd, "\r\n>-----------------[");
+                            wrstr(c, tls_fd, c_buf);
+                            wrstr(c, tls_fd, "]-----------------<\r\n\r\n");
+#else
+                            wrstr(c, tls_fd, "\r\n\r\n");
+                            wrstr(c, tls_fd, c_buf);
+                            wrstr(c, tls_fd, ":\r\n\r\n");
+#endif
+#ifdef WITH_RFC2640                            
+                            free(c_buf);
 #endif
                         }
                         if (!chdir(*path)) {
-                            listdir(0U, c, *path);
+                            listdir(0U, c, tls_fd, *path);
                             chdir(wd);
                         }
                     }
@@ -890,12 +960,15 @@ void donlist(char *arg, const int on_ctrl_conn, const int opt_l_,
         if (opt_d) {
             listfile(NULL, ".");
         } else {
-            listdir(0U, c, ".");
+            listdir(0U, c, tls_fd, ".");
         }
-        outputfiles(c);
+        outputfiles(c, tls_fd);
     }
-    wrstr(c, NULL);
+    wrstr(c, tls_fd, NULL);
     if (on_ctrl_conn == 0) {
+#ifdef WITH_TLS
+        closedata();
+#endif    
         close(c);
     } else {
         addreply_noformat(213, "End.");
